[
    {
        "ParameterKey": "AvailabilityZones",
        "ParameterValue": "$[taskcat_genaz_3]"
    },
    {
        "ParameterKey": "StackName",
        "ParameterValue": "TestStack"
    },
    {
        "ParameterKey": "ByteValue",
        "ParameterValue": 1
    },
    {
        "ParameterKey": "UUID",
        "ParameterValue": "$[taskcat_genuuid]"
    },
    {
        "ParameterKey": "RandomNumber",
        "ParameterValue": "$[taskcat_random-numbers]"
    },
    {
        "ParameterKey": "RandomString",
        "ParameterValue": "$[taskcat_random-string]"
    },
    {
        "ParameterKey": "UUID",
        "ParameterValue": "$[taskcat_genuuid]"
    },
    {
        "ParameterKey": "PasswordA",
        "ParameterValue": "$[taskcat_genpass_8A]"
    },
    {
        "ParameterKey": "PasswordAConfirm",
        "ParameterValue": "$[taskcat_getval_PasswordA]"
    },
    {
        "ParameterKey": "PasswordB",
        "ParameterValue": "$[taskcat_genpass_32S]"
    }
]
